from typing import Union
from shulker.functions.base_functions import *

############### BOSSBAR ###############

def meta_add_bossbar(name: str) -> dict:
    instructions = {"list": []}
    instructions["list"].append(f"bossbar add {id} {{\"text\":\"\"}}")
    return instructions

def meta_create_bossbar(id: str, text:str, target: str, value: int, color: str, progress: str, max: int, visible: bool) -> dict:
    instructions = {"list": []}
    
    instructions["list"].append(f"bossbar set {id} value {value}")
    instructions["list"].append(f"bossbar set {id} name {{\"text\":\"{text}\"}}")
    instructions["list"].append(f"bossbar set {id} color {color}")
    instructions["list"].append(f"bossbar set {id} players {target}")
    instructions["list"].append(f"bossbar set {id} style {progress}")
    instructions["list"].append(f"bossbar set {id} max {max}")
    instructions["list"].append(f"bossbar set {id} visible {str(visible).lower()}")    
    
    return instructions

def meta_list_bossbar():
    instructions = {"list": []}
    instructions["list"].append("bossbar list")
    return instructions

def meta_remove_bossbar():
    instructions = {"list": []}
    instructions["list"].append("bossbar remove")
    return instructions

def meta_get_bossbar(id: str, option: str):
    instructions = {"list": []}
    instructions["list"].append(f"bossbar get {id} {option}")
    return instructions

def meta_set_bossbar(id: str, option: str, value: str):
    instructions = {"list": []}
    instructions["list"].append(f"bossbar set {id} {option} {value}")
    return instructions

def add_bossbar(id: str) -> Union[bool, str]:

    check_output_channel()
    
    instructions = meta_add_bossbar(id)
    
    for line in instructions["list"]:
        status = post(line)
 
def create_bossbar(
    id: str,
    text: str,
    target: Union[str, None] = "@a",
    value: int = 100,
    color: str = "white",
    style: str = "progress",
    max: int = 100,
    visible: bool = True,
) -> Union[bool, str]:
    """
    Available colors: ["pink", "blue", "red", "green", "yellow", "purple", "white"]
    If target is None, it will only add the bossbar, but not display it to anyone.
    """

    check_output_channel()
    
    if target == None:
        target = ""
    
    instructions = meta_add_bossbar(id)
    instructions2 = meta_create_bossbar(id, text, target, value, color, style, max, visible)
    instructions["list"].extend(instructions2["list"])

    for line in instructions["list"]:
        status = post(line)

def list_bossbar() -> list:
    check_output_channel()
    
    instructions = meta_list_bossbar()
    
    for line in instructions["list"]:
        status = post(line)
        
    status = status[:-4]
    
    result = []
    bossbars = status.split(": ")[1].split(", ")
    for bossbar in bossbars:
        bossbar = bossbar.replace("[", "").replace("]", "")
        result.append(bossbar)
   
    return result

def remove_bossbar(id: str) -> Union[bool, str]:
    """
    Removes the bossbar with the given id
    """

    check_output_channel()
    
    instructions = meta_remove_bossbar()
    
    for line in instructions["list"]:
        status = post(line)
        
    if status.startswith("Removed"):
        return True
    else:
        return False

def get_bossbar(id: str, option: str) -> Union[bool, int]:
    """
    Returns a the values that were asked for corresponding to the bossbar id
    'players' options returns a list
    """

    check_output_channel()
    
    instructions = meta_get_bossbar(id, option)
    
    for line in instructions["list"]:
        status = post(line)[:-4]
    
    if "players" in option:
        value = status.split(": ")[1].split(", ")
    else:
        value = status.split(" ")[-1]
        
    if "<--[HERE]" in value:
        raise ValueError(f"bossbar get cannot fetch {option}'s value")
    
    return value

def set_bossbar(id: str, option: str, value: str) -> Union[bool, str]:
    """
    Sets the value of the bossbar with the given id
    Availables options: ["value", "max", "color", "style", "players", "name", "visible"]
    """

    check_output_channel()
    
    instructions = meta_set_bossbar(id, option, value)
    
    for line in instructions["list"]:
        status = post(line)
        
    if status.startswith("Set"):
        return True
    else:
        return False

############### TITLES ###############

def meta_show_gui(type: str, text: str, target: str) -> dict:
    instructions = {"list": []}
    instructions["list"].append(f"title {target} {type} {{\"text\":\"{text}\"}}")
    return instructions

def meta_clear_gui(target: str) -> dict:
    instructions = {"list": []}
    instructions["list"].append(f"title {target} clear")
    return instructions

def meta_set_gui_time(target: str, fade_in: int, stay: int, fade_out: int) -> dict:
    instructions = {"list": []}
    instructions["list"].append(f"title {target} times {fade_in} {stay} {fade_out}")
    return instructions

def show_gui(type: str, text: str, target: str = "@a") -> Union[bool, str]:
    """
    Shows a title or a subtitle or an actionbar to the given target
    Available types: ["title", "subtitle", "actionbar"]
    """

    check_output_channel()
    
    instructions = meta_show_gui(type, text, target)
    
    for line in instructions["list"]:
        status = post(line)

def clear_gui(target: str = "@a") -> Union[bool, str]:

    check_output_channel()
    
    instructions = {"list": []}
    instructions["list"].append(f"title {target} {type} clear")
    
    for line in instructions["list"]:
        status = post(line)
        
def set_gui_time(target: str = "@a", fade_in: int = 10, stay: int = 70, fade_out: int = 20) -> Union[bool, str]:
    """
    Sets the time of the title, subtitle or actionbar for the target
    Providing no values defaults to default time values
    """

    check_output_channel()
    
    instructions = meta_set_gui_time(target, fade_in, stay, fade_out)
    
    for line in instructions["list"]:
        status = post(line)


from typing import Union
from shulker.functions.base_functions import *
from shulker.components.Coordinates import Coordinates
import re

############ SAY ############

def meta_say(text: str) -> dict:
    instructions = {"list": []}
    instructions["list"].append(f"say {text}")
    return instructions

def say(text: str) -> Union[bool, str]:
    """This function sends a message in the chat"""
    check_output_channel()
    
    instructions = meta_say(text)
    
    for line in instructions["list"]:
        status = post(line)

######## BAN & KICK ########

def meta_ban(target: str, reason: str) -> dict:
    instructions = {"list": []}
    if reason is None:
        reason = ""
    instructions["list"].append(f"ban {target} {reason}")
    return instructions

def meta_banlist(option: str) -> dict:
    instructions = {"list": []}
    instructions["list"].append(f"banlist {option}")
    return instructions

def ban(target: str, reason: Union[str, None]) -> Union[bool, str]:
    """This function bans a player"""
    check_output_channel()
    
    instructions = meta_ban(target, reason)
    
    for line in instructions["list"]:
        status = post(line)
 
def ban_ip(target: str, reason: str) -> Union[bool, str]:
    """This function bans an IP"""
    check_output_channel()
    
    instructions = meta_ban(target, reason)
    
    for line in instructions["list"]:
        status = post(line)
        
def banlist(option: str) -> Union[bool, str]:
    """This functions fetches the banlist"""
    check_output_channel()
    
    instructions = meta_banlist(option)
    
    for line in instructions["list"]:
        status = post(line)

def meta_kick(target: str, reason: str) -> dict:
    instructions = {"list": []}
    if reason is None:
        reason = ""
    instructions["list"].append(f"kick {target} {reason}")
    return instructions

def kick(target: str, reason: Union[str, None]) -> Union[bool, str]:
    """This function kicks a player"""
    check_output_channel()
    
    instructions = meta_kick(target, reason)
    
    for line in instructions["list"]:
        status = post(line)

def meta_pardon(target: str) -> dict:
    instructions = {"list": []}
    instructions["list"].append(f"pardon {target}")
    return instructions

def meta_pardon_ip(target: str) -> dict:
    instructions = {"list": []}
    instructions["list"].append(f"pardon-ip {target}")
    return instructions

def pardon(target: str) -> Union[bool, str]:
    """This function unbans a player"""
    check_output_channel()
    
    instructions = meta_pardon(target)
    
    for line in instructions["list"]:
        status = post(line)

def pardon_ip(target: str) -> Union[bool, str]:
    """This function unbans an IP"""
    check_output_channel()
    
    instructions = meta_pardon_ip(target)
    
    for line in instructions["list"]:
        status = post(line)

############ OP ############

def meta_op(target: str) -> dict:
    instructions = {"list": []}
    instructions["list"].append(f"op {target}")
    return instructions

def meta_deop(target: str) -> dict:
    instructions = {"list": []}
    instructions["list"].append(f"deop {target}")
    return instructions

def op(target: str) -> Union[bool, str]:
    """This function gives an operator status to a player"""
    check_output_channel()
    
    instructions = meta_op(target)
    
    for line in instructions["list"]:
        status = post(line)
        
def deop(target: str) -> Union[bool, str]:
    """This function removes an operator status from a player"""
    check_output_channel()
    
    instructions = meta_deop(target)
    
    for line in instructions["list"]:
        status = post(line)
        
############# SEED #############

def meta_seed() -> dict:
    instructions = {"list": []}
    instructions["list"].append(f"seed")
    return instructions

def seed() -> Union[bool, str]:
    """This function fetches the seed of the world"""
    check_output_channel()
    
    instructions = meta_seed()
    
    for line in instructions["list"]:
        status = post(line)
    
    seed = re.match(r"Seed: \[(-{0,1}(\d+))\]", status)
    return seed.group(1)

############# DIFFICULTY #############

def meta_set_difficulty(option: str) -> dict:
    instructions = {"list": []}
    instructions["list"].append(f"difficulty {option}")
    return instructions

def meta_get_difficulty() -> dict:
    instructions = {"list": []}
    instructions["list"].append(f"difficulty")
    return instructions

def set_difficulty(option: str) -> Union[bool, str]:
    """This function sets the difficulty of the world"""
    check_output_channel()
    
    instructions = meta_set_difficulty(option)
    
    for line in instructions["list"]:
        status = post(line)
        
def get_difficulty() -> Union[bool, str]:
    """This function fetches the difficulty of the world"""
    check_output_channel()
    
    instructions = meta_get_difficulty()
    
    for line in instructions["list"]:
        status = post(line)
    
    difficulty = re.match(r"The difficulty is (\w+)", status)
    return difficulty.group(1)


########### WEATHER ###########

def meta_weather(option: str) -> dict:
    instructions = {"list": []}
    instructions["list"].append(f"weather {option}")
    return instructions

def weather(option: str) -> Union[bool, str]:
    """This function sets the weather of the world
    Options: clear, rain, thunder"""
    
    check_output_channel()
    
    instructions = meta_weather(option)
    
    for line in instructions["list"]:
        status = post(line)
        
########### MSG ###########

def meta_msg(target: str, message: str) -> dict:
    instructions = {"list": []}
    instructions["list"].append(f"msg {target} {message}")
    return instructions

def msg(target: str, message: str) -> Union[bool, str]:
    """This function sends a private message to a player"""
    
    check_output_channel()
    
    instructions = meta_msg(target, message)
    
    for line in instructions["list"]:
        status = post(line)

########### GAMEMODE ###########

def meta_default_gamemode(option: str) -> dict:
    instructions = {"list": []}
    instructions["list"].append(f"defaultgamemode {option}")
    return instructions

def meta_gamemode(target: str, option: str) -> dict:
    instructions = {"list": []}
    instructions["list"].append(f"gamemode {option} {target}")
    return instructions

def gamemode(target: str, option: str) -> Union[bool, str]:
    """This function sets the gamemode of a player
    Available options: survival, creative, adventure, spectator"""
    
    check_output_channel()
    
    instructions = meta_gamemode(target, option)
    
    for line in instructions["list"]:
        status = post(line)
        
def default_gamemode(option: str) -> Union[bool, str]:
    """This function sets the default gamemode of the world
    Available options: survival, creative, adventure, spectator"""
    
    check_output_channel()
    
    instructions = meta_default_gamemode(option)
    
    for line in instructions["list"]:
        status = post(line)

########### TIME ###########

def meta_query_time(option: str) -> dict:
    instructions = {"list": []}
    instructions["list"].append(f"time query {option}")
    return instructions

def query_time(option: str = "day") -> Union[bool, str]:
    """This function fetches the time of the world
    Available options: day, daytime, gametime"""
    
    check_output_channel()
    
    instructions = meta_query_time(option)
    
    for line in instructions["list"]:
        status = post(line)
        
    time = re.match(r"The time is (\d+)", status)
    return time.group(1)

def meta_add_time(value: int, option: str) -> dict:
    instructions = {"list": []}
    instructions["list"].append(f"time add {value}{option}")
    return instructions

def add_time(value: int, option: str = "tick") -> Union[bool, str]:
    """This function adds time to the world
    Available options: day, second, tick"""
    
    check_output_channel()
    
    option = option[0]
    
    instructions = meta_add_time(value, option)
    
    for line in instructions["list"]:
        status = post(line)
        
def meta_set_time(value: int, option: str) -> dict:
    instructions = {"list": []}
    instructions["list"].append(f"time set {value}{option}")
    return instructions

def set_time(value: int, option: str = "tick") -> Union[bool, str]:
    """This function sets the time of the world
    Available options: day, second, tick"""
    
    check_output_channel()
    
    option = option[0]
    
    instructions = meta_set_time(value, option)
    
    for line in instructions["list"]:
        status = post(line)
        
def meta_time(option: str) -> dict:
    instructions = {"list": []}
    instructions["list"].append(f"time set {option}")
    return instructions

def time(option: str) -> Union[bool, str]:
    """This function sets the time of the world
    Available options: day, midnight, night, noon"""
    
    check_output_channel()
    
    instructions = meta_time(option)
    
    for line in instructions["list"]:
        status = post(line)
        
########### EXPERIENCE ###########

def meta_xp_query(target: str, option: str) -> dict:
    instructions = {"list": []}
    instructions["list"].append(f"xp query {target} {option}")
    return instructions

def xp_query(target: str, option: str) -> Union[bool, str]:
    """This function fetches the experience of a player
    Available options: levels, points"""
    
    check_output_channel()
    
    instructions = meta_xp_query(target, option)
    
    for line in instructions["list"]:
        status = post(line)
    
    xp = re.findall(r"(\d+) experience " + option, status)
    return xp[0]

########### WHITELIST ###########

def meta_get_whitelist() -> dict:
    instructions = {"list": []}
    instructions["list"].append(f"whitelist list")
    return instructions

def get_whitelist() -> Union[bool, str]:
    """This function fetches the whitelist of the world"""
    
    check_output_channel()
    
    instructions = meta_get_whitelist()
    
    for line in instructions["list"]:
        status = post(line)[:-4]
        
    whitelist = status.split("players: ")[1].split(", ")
    return whitelist

def meta_toggle_whitelist(option: str) -> dict:
    instructions = {"list": []}
    instructions["list"].append(f"whitelist {option}")
    return instructions

def toggle_whitelist(option: str) -> Union[bool, str]:
    """This function toggles the whitelist of the world
    Available options: on, off"""
    
    check_output_channel()
    
    instructions = meta_toggle_whitelist(option)
    
    for line in instructions["list"]:
        status = post(line)
        
def meta_reload_whitelist() -> dict:
    instructions = {"list": []}
    instructions["list"].append(f"whitelist reload")
    return instructions

def reload_whitelist() -> Union[bool, str]:
    """This function reloads the whitelist of the world"""
    
    check_output_channel()
    
    instructions = meta_reload_whitelist()
    
    for line in instructions["list"]:
        status = post(line)
        
def meta_update_whitelist(option: str, target: str) -> dict:
    instructions = {"list": []}
    instructions["list"].append(f"whitelist {option} {target}")
    return instructions

def update_whitelist(option: str, target: str) -> Union[bool, str]:
    """This function updates the whitelist of the world
    Available options: add, remove"""
    
    check_output_channel()
    
    instructions = meta_update_whitelist(option, target)
    
    for line in instructions["list"]:
        status = post(line)
        
############ ADMIN ############

def meta_stop() -> dict:
    instructions = {"list": []}
    instructions["list"].append(f"stop")
    return instructions

def stop() -> Union[bool, str]:
    """This function stops the server"""
    
    check_output_channel()
    
    instructions = meta_stop()
    
    for line in instructions["list"]:
        status = post(line)
        
def meta_save_all() -> dict:
    instructions = {"list": []}
    instructions["list"].append(f"save-all")
    return instructions

def save_all() -> Union[bool, str]:
    """This function saves the world"""
    
    check_output_channel()
    
    instructions = meta_save_all()
    
    for line in instructions["list"]:
        status = post(line)
        
def meta_toggle_save(option: str) -> dict:
    instructions = {"list": []}
    instructions["list"].append(f"save-{option}")
    return instructions

def toggle_save(option: str) -> Union[bool, str]:
    """This function toggles the auto-save of the world
    Available options: on, off"""
    
    check_output_channel()
    
    instructions = meta_toggle_save(option)
    
    for line in instructions["list"]:
        status = post(line)
        
############# HELP #############

def meta_help(value: str) -> dict:
    instructions = {"list": []}
    instructions["list"].append(f"help {value}")
    return instructions

def parse_help(prepender = ""):
    
    instructions = meta_help(prepender)
    
    for line in instructions["list"]:
        data = post(line)
        
    # Remove all \x1b[.. color code from the status
    data = re.sub(r"\x1b\[[0-9;]*m", "", data)
    
    print(data)
    # Find the number of help pages
    nb_pages = re.findall(r" \(\d+/(\d+)\)", data)
    
    if nb_pages:
        nb_pages = nb_pages[0]
    elif " Help: " in data and not nb_pages:
        nb_pages = 1
    else:
        raise Exception("No help found")
        
    help_cmds = {}
    for page in range(1, int(nb_pages) + 1):
                
        instructions = meta_help(prepender + str(page))
        
        for line in instructions["list"]:
            data = post(line)
            
        data = re.sub(r"\x1b\[[0-9;]*m", "", data)
        lines = data.splitlines()[(1 if page > 1 else 2):]
        
        for line in lines:
            if not line.startswith("/") and ":" not in line:
                continue
            splitted = line.split(": ")
            command = splitted[0]
            value = splitted[1]
            help_cmds[command] = value
    
    return help_cmds

def help(value: str = "") -> Union[bool, dict]:
    """This function fetches the help of the world"""
    
    check_output_channel()

    cmds = parse_help(value)
    for key in cmds:
        if not key.startswith('/'):
            cmds[key] = parse_help(prepender=key + " ")
            
    return cmds

############# LIST #############

def meta_list_players(uuids: bool = False) -> dict:
    instructions = {"list": []}
    if uuids:
        uuids = "uuids"
    else:
        uuids = ""
    instructions["list"].append(f"list {uuids}")
    return instructions

def list_players(uuids: bool = False) -> Union[bool, str]:
    """This function fetches the list of players in the world
    If uuids is set to True it will be a list of uuids returned"""
    
    check_output_channel()
    
    instructions = meta_list_players(uuids)
    
    for line in instructions["list"]:
        status = post(line)
    
    # Remove all \x1b[.. color code from the status
    status = re.sub(r"\x1b\[[0-9;]*m", "", status)
    
    if uuids:
        players = status
    else:
        players = status.split("players online: ")[1].split(", ")
        
    return players

############# SPECTATE #############

def meta_spectate(target: str, player: str) -> dict:
    instructions = {"list": []}
    instructions["list"].append(f"spectate {target} {player}")
    return instructions

def spectate(target: str, player: str) -> Union[bool, str]:
    """This function makes the target spectates a player"""
    
    check_output_channel()
    
    instructions = meta_spectate(target, player)
    
    for line in instructions["list"]:
        status = post(line)
        
        
############# SET WORLD SPAWN #############

def meta_set_world_spawn(pos: Coordinates, yaw: int) -> dict:
    instructions = {"list": []}
    instructions["list"].append(f"setworldspawn {pos} {yaw}")
    return instructions

def set_world_spawn(pos: Union[Coordinates, tuple], yaw: int = 0) -> Union[bool, str]:
    """This function sets the world spawn"""
    
    check_output_channel()
    
    pos = format_arg(pos, Coordinates)

    instructions = meta_set_world_spawn(pos, yaw)
    
    for line in instructions["list"]:
        status = post(line)


from .base_functions import nest_commands, entity_list
from .get_block import get_block, meta_get_block
from .set_block import set_block, meta_set_block
from .set_image import set_image, meta_set_image, print_palette
from .set_text import set_text, meta_set_text
from .set_zone import set_zone, meta_set_zone
from .summon import summon, meta_summon
from .get_player_nbt import get_player_nbt, get_player_pos, meta_get_player_nbt

from .set_gui import create_bossbar, meta_create_bossbar
from .set_gui import add_bossbar, meta_add_bossbar
from .set_gui import list_bossbar, meta_list_bossbar
from .set_gui import remove_bossbar, meta_remove_bossbar
from .set_gui import get_bossbar, meta_get_bossbar
from .set_gui import set_bossbar, meta_set_bossbar
from .set_gui import show_gui, meta_show_gui
from .set_gui import clear_gui, meta_clear_gui

from .miscellaneous import say, meta_say
from .miscellaneous import ban, ban_ip, meta_ban, banlist, meta_banlist, kick, meta_kick, pardon, meta_pardon, pardon_ip, meta_pardon_ip
from .miscellaneous import op, deop, meta_op, meta_deop
from .miscellaneous import seed, meta_seed
from .miscellaneous import set_difficulty, meta_set_difficulty, get_difficulty, meta_get_difficulty
from .miscellaneous import weather, meta_weather
from .miscellaneous import msg, meta_msg
from .miscellaneous import gamemode, meta_gamemode, default_gamemode, meta_default_gamemode
from .miscellaneous import query_time, meta_query_time, set_time, meta_set_time, add_time, meta_add_time, time, meta_time
from .miscellaneous import xp_query, meta_xp_query
from .miscellaneous import get_whitelist, meta_get_whitelist, toggle_whitelist, meta_toggle_whitelist, update_whitelist, meta_update_whitelist
from .miscellaneous import stop, meta_stop, save_all, meta_save_all, toggle_save, meta_toggle_save
from .miscellaneous import help, meta_help
from .miscellaneous import list_players, meta_list_players
from .miscellaneous import spectate, meta_spectate
from .miscellaneous import set_world_spawn, meta_set_world_spawn


import re
from typing import Union

from shulker.components.BlockCoordinates import BlockCoordinates
from shulker.components.Coordinates import Coordinates
from shulker.components.NBT import NBT

from shulker.functions.base_functions import *

def meta_get_player_nbt(pseudo: str) -> dict:
    instructions = {"list": []}
    instructions["list"].append(f"data get entity {pseudo}")
    return instructions


def get_player_nbt(
    pseudo: str,
) -> Union[bool, str]:
    """
    Returns the NBT data of the player
    """

    check_output_channel()
    
    if type(pseudo) is not str:
        raise TypeError(f"Expected type str, got {type(pseudo)}")

    instructions = meta_get_player_nbt(pseudo)

    for line in instructions["list"]:
        status = post(line)
        
    if status.startswith(f"{pseudo} has"):
        data = status.split("data: ")[1][:-4]
        return NBT(data)
    else:
        return False
    

def get_player_pos(
    pseudo: str,
    rounded: bool = True
) -> Union[bool, str]:
    """
    Returns the coordinates of the player if it is found
    False if it wasn't found or there's an issue
    """

    check_output_channel()
    
    nbt = get_player_nbt(pseudo)
    
    if hasattr(nbt, "Pos"):
        pos = nbt.Pos
        if rounded:
            return BlockCoordinates(pos[0], pos[1], pos[2])
        else:
            return Coordinates(float(pos[0]), float(pos[1]), float(pos[2]))
    else:
        return False
